@@ -4,7 +4,7 @@ use strict;
 use warnings;
 
 our $VERSION;
-$VERSION = '0.07';
+$VERSION = '0.08';
 
 use Net::DNS qw(:DEFAULT);
 use base     qw(Exporter Net::DNS);
@@ -13,15 +13,22 @@ our @EXPORT = @Net::DNS::EXPORT;
 
 use IO::Select;
 use IO::Socket;
-use Socket qw(pack_ip_mreq pack_ipv6_mreq inet_pton
-		IPPROTO_IP IPPROTO_IPV6 IP_MULTICAST_TTL IPV6_MULTICAST_HOPS
-		IP_ADD_MEMBERSHIP IPV6_ADD_MEMBERSHIP IPV6_JOIN_GROUP);
+use Socket qw(	pack_ipv6_mreq inet_pton
+		IPPROTO_IP IP_MULTICAST_TTL
+		IPPROTO_IPV6 IPV6_MULTICAST_HOPS
+		IPV6_ADD_MEMBERSHIP IPV6_JOIN_GROUP IPV6_MULTICAST_LOOP);
 
 use constant REUSEPORT => defined eval {
 	my $socket = Net::DNS::Resolver->new->_create_udp_socket('224.0.0.251');
 	$socket->setsockopt( SOL_SOCKET, SO_REUSEPORT, 1 );
 };
 
+use constant AVOIDLOOP => defined eval {
+	my $socket6 = Net::DNS::Resolver->new->_create_udp_socket('FF02::FB');
+	my $ip_mreq = pack_ipv6_mreq( inet_pton( AF_INET6, 'FF02::FB' ), 0 );
+	$socket6->setsockopt( IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 0 );
+};
+
 use constant ADDMEMBER => defined eval {
 	my $socket6 = Net::DNS::Resolver->new->_create_udp_socket('FF02::FB');
 	my $ip_mreq = pack_ipv6_mreq( inet_pton( AF_INET6, 'FF02::FB' ), 0 );
@@ -45,7 +52,11 @@ Net::DNS::Multicast - Multicast extension to Net::DNS
     use Net::DNS::Multicast;
     my $resolver = Net::DNS::Resolver->new();
     my $response = $resolver->send( 'host.local.', 'AAAA' );
-    my $discover = $resolver->send( '_ipp._tcp.local.', 'PTR' );
+
+    my $handle = $resolver->bgsend( '_ipp._tcp.local.', 'PTR' );
+    while ( my $response = $resolver->bgread($handle) ) {
+	$response->print;
+    }
 
 =head1 DESCRIPTION
 
@@ -92,30 +103,39 @@ sub Net::DNS::Resolver::bgsend {
 	return IO::Select->new( Net::DNS::Resolver::Base::bgsend( $self, $packet ) )
 			unless $q->qname =~ /$NAME_REGEX/oi;
 
-	$self->_reset_errorstring;
-	local $packet->{status} = 0;
 	local @{$self}{qw(nameservers nameserver4 nameserver6 port)};
-	my $select = IO::Select->new();
 	my $port   = $self->port($multicast_port);
+	my $select = IO::Select->new();
+	my $expire = time() + $multicast_timeout;
+	$packet->{status} = 0;
+	$self->_reset_errorstring;
 	foreach my $ip ( $self->nameservers(@multicast_group) ) {
 		my $socket = $self->_create_udp_socket($ip);
 		next unless $socket;				# uncoverable branch true
-		$socket->setsockopt( SOL_SOCKET, SO_REUSEADDR, 1 );
-		$socket->setsockopt( SOL_SOCKET, SO_REUSEPORT, 1 ) if REUSEPORT;
 		if ( $socket->sockdomain() == AF_INET6 ) {
 			my $ip_mreq = pack_ipv6_mreq( inet_pton( AF_INET6, $ip ), 0 );
-			$socket->setsockopt( IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, $ip_mreq ) if ADDMEMBER;
-			$socket->setsockopt( IPPROTO_IPV6, IPV6_JOIN_GROUP,	$ip_mreq ) if JOINGROUP;
+			my @sockopt = ( [SOL_SOCKET, SO_REUSEADDR, 1] );
+			push @sockopt, [SOL_SOCKET,   SO_REUSEPORT,	   1]	     if REUSEPORT;
+			push @sockopt, [IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 0]	     if AVOIDLOOP;
+			push @sockopt, [IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, $ip_mreq] if ADDMEMBER;
+			push @sockopt, [IPPROTO_IPV6, IPV6_JOIN_GROUP,	   $ip_mreq] if JOINGROUP;
+
+			my $listener = $self->_create_udp_socket(
+				$ip,
+				LocalPort => $port,
+				Sockopts  => [@sockopt],
+				);
+			if ($listener) {			# uncoverable branch false
+				${*$listener}{net_dns_bg} = [$expire, $packet];
+				$select->add($listener);	# multicast response, if any
+			}
 			$socket->setsockopt( IPPROTO_IPV6, IPV6_MULTICAST_HOPS, 1 );
 		} else {
-			my $ip_mreq = pack_ip_mreq( inet_aton($ip), INADDR_ANY );
-			$socket->setsockopt( IPPROTO_IP, IP_ADD_MEMBERSHIP, $ip_mreq );
-			$socket->setsockopt( IPPROTO_IP, IP_MULTICAST_TTL,  1 );
+			$socket->setsockopt( IPPROTO_IP, IP_MULTICAST_TTL, 1 );
 		}
 
-		my $expire = time() + $multicast_timeout;
-		${*$socket}{net_dns_bg} = [$expire];
-		$select->add($socket);
+		${*$socket}{net_dns_bg} = [$expire, $packet];
+		$select->add($socket);				# unicast response, if any
 
 		$self->_diag( 'bgsend', "[$ip]:$port" );
 		my $sockaddr = $self->_create_dst_sockaddr( $ip, $port );
@@ -135,8 +155,11 @@ sub Net::DNS::Resolver::bgbusy {
 
 sub Net::DNS::Resolver::bgread {
 	my ( $self, $select ) = @_;
-	my ($handle) = ( $select->can_read(0), $select->handles );
-	return Net::DNS::Resolver::Base::bgread( $self, $handle );
+	my $response;
+	foreach my $handle ( $select->can_read(0), $select->handles ) {
+		last if $response = Net::DNS::Resolver::Base::bgread( $self, $handle );
+	}
+	return $response;
 }
 
 
